"Utilities for implementing operators."


def extend_name(suffix):
    """A factory for class decorators that modify the class name by appending some text to it.

    Example:

    .. code-block:: python

        @extend_name('_Foo')
        class Class:
            pass

        assert Class.__name__ == 'Class_Foo'
    """

    def dec(cls):
        name = "{}{}".format(cls.__name__, suffix)
        setattr(cls, "__name__", name)
        return cls

    return dec



"A tool for creating XML reports."

import sys
import xml.etree.ElementTree

import click

from cosmic_ray.work_db import WorkDB, use_db
from cosmic_ray.work_item import TestOutcome, WorkerOutcome


@click.command()
@click.argument("session-file", type=click.Path(dir_okay=False, readable=True, exists=True))
def report_xml(session_file):
    """Print an XML formatted report of test results for continuous integration systems"""
    with use_db(session_file, WorkDB.Mode.open) as db:
        xml_elem = _create_xml_report(db)
        xml_elem.write(sys.stdout.buffer, encoding="utf-8", xml_declaration=True)


def _create_xml_report(db):
    errors = 0
    failed = 0
    skipped = 0
    root_elem = xml.etree.ElementTree.Element("testsuite")

    for work_item, result in db.completed_work_items:
        if result.worker_outcome in {WorkerOutcome.EXCEPTION, WorkerOutcome.ABNORMAL}:
            errors += 1
        if result.is_killed:
            failed += 1
        if result.worker_outcome == WorkerOutcome.SKIPPED:
            skipped += 1

        subelement = _create_element_from_work_item(work_item)
        subelement = _update_element_with_result(subelement, result)
        root_elem.append(subelement)

    for work_item in db.pending_work_items:
        subelement = _create_element_from_work_item(work_item)
        root_elem.append(subelement)

    root_elem.set("errors", str(errors))
    root_elem.set("failures", str(failed))
    root_elem.set("skips", str(skipped))
    root_elem.set("tests", str(db.num_work_items))
    return xml.etree.ElementTree.ElementTree(root_elem)


def _create_element_from_work_item(work_item):
    sub_elem = xml.etree.ElementTree.Element("testcase")

    for mutation in work_item.mutations:
        mutation_elem = xml.etree.ElementTree.Element("mutation")
        mutation_elem.set("classname", work_item.job_id)
        mutation_elem.set("line", str(mutation.start_pos[0]))
        mutation_elem.set("file", str(mutation.module_path))
        sub_elem.append(mutation_elem)

    return sub_elem


def _update_element_with_result(sub_elem, result):
    data = result.output
    outcome = result.worker_outcome

    if outcome == WorkerOutcome.EXCEPTION:
        error_elem = xml.etree.ElementTree.SubElement(sub_elem, "error")
        error_elem.set("message", "Worker has encountered exception")
        error_elem.text = str(data) + "\n".join(result.diff)
    elif _evaluation_success(result):
        failure_elem = xml.etree.ElementTree.SubElement(sub_elem, "failure")
        failure_elem.set("message", "Mutant has survived your unit tests")
        failure_elem.text = str(data) + result.diff

    return sub_elem


def _evaluation_success(result):
    return result.worker_outcome == WorkerOutcome.NORMAL and result.test_outcome in {
        TestOutcome.SURVIVED,
        TestOutcome.INCOMPETENT,
    }


if __name__ == "__main__":
    report_xml()  # no-qa: no-value-for-parameter


"""A simple base for creating common types of work-db filters.
"""
import argparse
import logging
import sys

from exit_codes import ExitCode

from cosmic_ray.work_db import use_db


class FilterApp:
    """Base class for simple WorkDB filters.

    This provides command-line handling for common filter options like
    the session and verbosity level. Subclasses can add their own arguments
    as well. This provides a `main()` function that open the session's WorkDB
    and passes it to the subclass's `filter()` function.
    """

    def add_args(self, parser: argparse.ArgumentParser):
        """Add any arguments that the subclass needs to the parser.

        Args:
            parser: The ArgumentParser for command-line processing.
        """

    def description(self):
        """The description of the filter.

        This is used for the command-line help message.
        """
        return None

    def main(self, argv=None):
        """The main function for the app.

        Args:
            argv: Command line argument list of parse.
        """
        if argv is None:
            argv = sys.argv[1:]

        parser = argparse.ArgumentParser(
            description=self.description(),
        )
        parser.add_argument(
            'session', help="Path to the session on which to operate")
        parser.add_argument(
            '--verbosity', help='Verbosity level for logging', default='WARNING')
        self.add_args(parser)
        args = parser.parse_args(argv)

        logging.basicConfig(level=getattr(logging, args.verbosity))

        with use_db(args.session) as db:
            self.filter(db, args)

        return ExitCode.OK

    def filter(self, work_db, args):
        """Apply this filter to a WorkDB.

        This should modify the WorkDB in place.

        Args:
            work_db: An open WorkDB instance.
            args: The argparse Namespace for the command line.
        """
        raise NotImplementedError()



"Tool for printing the survival rate in a session."

import math
import sys

import click

from cosmic_ray.work_db import WorkDB, use_db

SUPPORTED_Z_SCORES = {800: 1.282, 900: 1.645, 950: 1.960, 980: 2.326, 990: 2.576, 995: 2.807, 998: 3.080, 999: 3.291}


@click.command()
@click.option(
    "--estimate/--no-estimate", default=False, help="Print the lower bound, estimate and upper bound of survival rate"
)
@click.option(
    "--confidence",
    type=click.Choice(sorted([str(z / 10) for z in SUPPORTED_Z_SCORES])),
    default=95,
    help="Specify the confidence levels for estimates",
)
@click.option(
    "--fail-over",
    type=click.FloatRange(0, 100),
    default=None,
    help="Exit with a non-zero code if the survival rate is larger than <max_value> or the calculated confidence interval is above the <max_value> (if --estimate is used).  Specified as percentage.",
)
@click.argument("session-file", type=click.Path(dir_okay=False, readable=True, exists=True))
def format_survival_rate(estimate, confidence, fail_over, session_file):
    """Calculate the survival rate of a session."""
    confidence = float(confidence)
    try:
        z_score = SUPPORTED_Z_SCORES[int(float(confidence) * 10)]
    except KeyError:
        raise ValueError("Unsupported confidence interval: {0}".format(confidence))

    with use_db(session_file, WorkDB.Mode.open) as db:
        rate = survival_rate(db)
        num_items = db.num_work_items
        num_complete = db.num_results

    if estimate:
        conf_int = math.sqrt(rate * (100 - rate) / num_complete) * z_score * (1 - math.sqrt(num_complete / num_items))
        min_rate = rate - conf_int
        print("{:.2f} {:.2f} {:.2f}".format(min_rate, rate, rate + conf_int))

    else:
        print("{:.2f}".format(rate))
        min_rate = rate

    if fail_over and min_rate > float(fail_over):
        sys.exit(1)


def kills_count(work_db):
    """Return the number of killed mutants."""
    return sum(r.is_killed for _, r in work_db.results)


def survival_rate(work_db):
    """Calculate the survival rate for the results in a WorkDB."""
    kills = kills_count(work_db)
    num_results = work_db.num_results

    if not num_results:
        return 0

    return (1 - kills / num_results) * 100



"""Configuration module."""
from contextlib import contextmanager
import logging
import sys

import toml

log = logging.getLogger()


def load_config(filename=None):
    """Load a configuration from a file or stdin.

    If `filename` is `None` or "-", then configuration gets read from stdin.

    Returns: A `ConfigDict`.

    Raises: ConfigError: If there is an error loading the config.
    """
    try:
        with _config_stream(filename) as handle:
            filename = handle.name
            return deserialize_config(handle.read())
    except (OSError, toml.TomlDecodeError, UnicodeDecodeError) as exc:
        raise ConfigError("Error loading configuration from {}".format(filename)) from exc


def deserialize_config(sz) -> "ConfigDict":
    "Parse a serialized config into a ConfigDict."
    return toml.loads(sz, _dict=ConfigDict)["cosmic-ray"]


def serialize_config(config):
    "Return the serialized form of `config`."
    return toml.dumps({"cosmic-ray": config})


class ConfigError(Exception):
    "Base class for exceptions raised by ConfigDict."


class ConfigKeyError(ConfigError, KeyError):
    "KeyError subclass raised by ConfigDict."


class ConfigValueError(ConfigError, ValueError):
    "ValueError subclass raised by ConfigDict."


class ConfigDict(dict):
    """A dictionary subclass that contains the application configuration."""

    def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        except KeyError as exc:
            raise ConfigKeyError(*exc.args)

    def sub(self, *segments):
        "Get a sub-configuration."
        d = self
        for segment in segments:
            try:
                d = d[segment]
            except KeyError:
                return ConfigDict({})
        return d

    @property
    def test_command(self):
        """The command to run to execute tests."""
        return self["test-command"]

    @property
    def timeout(self):
        "The timeout (seconds) for tests."
        return float(self["timeout"])

    @property
    def distributor_name(self):
        "The name of the distributor to use."
        return self["distributor"]["name"]

    @property
    def distributor_config(self):
        "The configuration for the named distributor."
        name = self.distributor_name
        return self["distributor"].get(name, ConfigDict())


@contextmanager
def _config_stream(filename):
    """Given a configuration's filename, this returns a stream from which a configuration can be read.

    If `filename` is `None` or '-' then stream will be `sys.stdin`. Otherwise,
    it's the open file handle for the filename.
    """
    if filename is None or filename == "-":
        log.info("Reading config from stdin")
        yield sys.stdin
    else:
        with open(filename, mode="rt") as handle:
            log.info("Reading config from %r", filename)
            yield handle



"A tool for generating HTML reports."

import datetime

from itertools import chain

import click
from yattag import Doc

from cosmic_ray.work_db import WorkDB, use_db
from cosmic_ray.work_item import TestOutcome
from cosmic_ray.tools.survival_rate import kills_count, survival_rate


@click.command()
@click.option("--only-completed/--not-only-completed", default=False)
@click.option("--skip-success/--include-success", default=False)
@click.argument("session-file", type=click.Path(dir_okay=False, readable=True, exists=True))
def report_html(only_completed, skip_success, session_file):
    """Print an HTML formatted report of test results."""
    with use_db(session_file, WorkDB.Mode.open) as db:
        doc = _generate_html_report(db, only_completed, skip_success)

    print(doc.getvalue())


def _generate_html_report(db, only_completed, skip_success):
    # pylint: disable=too-many-statements
    doc, tag, text = Doc().tagtext()
    doc.asis("<!DOCTYPE html>")
    with tag("html", lang="en"):
        with tag("head"):
            doc.stag("meta", charset="utf-8")
            doc.stag("meta", name="viewport", content="width=device-width, initial-scale=1, shrink-to-fit=no")
            doc.stag(
                "link",
                rel="stylesheet",
                href="https://stackpath.bootstrapcdn.com/bootstrap/4.3.1/css/bootstrap.min.css",
                integrity="sha384-ggOyR0iXCbMQv3Xipma34MD+dH/1fQ784/j6cY/iJTQUOhcWr7x9JvoRxT2MZw1T",
                crossorigin="anonymous",
            )
            with tag("title"):
                text("Cosmic Ray Report")
        with tag("body"):
            with tag("div", klass="container"):
                with tag("h1"):
                    with tag("p", klass="text-dark"):
                        text("Cosmic Ray Report")

            all_items = db.completed_work_items
            if not only_completed:
                incomplete = ((item, None) for item in db.pending_work_items)
                all_items = chain(all_items, incomplete)

            num_items = db.num_work_items
            num_complete = db.num_results

            with tag("div", klass="container"):

                # Summary info

                with tag("div", klass="mb-1", id="summary_info___accordion"):
                    with tag("div", klass="card"):
                        with tag(
                            "a",
                            ("data-toggle", "collapse"),
                            ("data-target", "#summary_info___collapse_1"),
                            ("aria-expanded", "true"),
                            ("aria-controls", "summary_info___collapse_1"),
                            href="#",
                        ):
                            with tag("div", klass="card-header", id="summary_info___heading_1"):
                                with tag("button", klass="btn btn-outline-dark"):
                                    with tag("h4", klass="m-0"):
                                        text("Summary info")

                        with tag(
                            "div",
                            ("aria-labelledby", "summary_info___heading_1"),
                            ("data-parent", "#summary_info___accordion"),
                            klass="collapse show",
                            id="summary_info___collapse_1",
                        ):
                            with tag("div", klass="card-body"):
                                with tag("p"):
                                    text("Date time: {}".format(datetime.datetime.now().strftime("%d/%m/%Y %H:%M:%S")))

                                with tag("p"):
                                    text("Total jobs: {}".format(num_items))

                                if num_complete > 0:
                                    with tag("p"):
                                        text(
                                            "Complete: {} ({:.2f}%)".format(
                                                num_complete, num_complete / num_items * 100
                                            )
                                        )
                                    with tag("p"):
                                        num_killed = kills_count(db)
                                        text(
                                            "Surviving mutants: {} ({:.2f}%)".format(
                                                num_complete - num_killed, survival_rate(db)
                                            )
                                        )
                                else:
                                    with tag("p"):
                                        text("No jobs completed")

                # Job list

                with tag("div", klass="mb-1", id="job_list___accordion"):
                    with tag("div", klass="card"):
                        with tag(
                            "a",
                            ("data-toggle", "collapse"),
                            ("data-target", "#job_list___collapse_1"),
                            ("aria-expanded", "false"),
                            ("aria-controls", "job_list___collapse_1"),
                            href="#",
                        ):
                            with tag("div", klass="card-header", id="job_list___heading_1"):
                                with tag("button", klass="btn btn-outline-dark"):
                                    with tag("h4", klass="m-0"):
                                        text("Job list")

                        with tag(
                            "div",
                            ("aria-labelledby", "job_list___heading_1"),
                            ("data-parent", "#job_list___accordion"),
                            klass="collapse",
                            id="job_list___collapse_1",
                        ):
                            with tag("div", klass="card-body"):
                                with tag("div", klass="text-right mb-1"):
                                    with tag("div", klass="mx-1", id="job_item_expand_all"):
                                        with tag(
                                            "a",
                                            href="#",
                                            onclick="$('div.job_list___sub_multi_collapse').collapse('show');",
                                        ):
                                            with tag("button", klass="btn btn-outline-dark"):
                                                with tag("span"):
                                                    text("Expand All")
                                    with tag("div", klass="mx-1", id="job_item_collapse_all"):
                                        with tag(
                                            "a",
                                            href="#",
                                            onclick="$('div.job_list___sub_multi_collapse').collapse('hide');",
                                        ):
                                            with tag("button", klass="btn btn-outline-dark"):
                                                with tag("span"):
                                                    text("Collapse All")

                                # Job item

                                for index, (work_item, result) in enumerate(all_items, start=1):
                                    if result is not None:
                                        if result.is_killed:
                                            if result.test_outcome == TestOutcome.INCOMPETENT:
                                                level = "info"
                                            else:
                                                level = "success"
                                                if skip_success:
                                                    continue
                                        else:
                                            level = "danger"

                                    with tag("div", klass="mb-1", id="job_list___sub_accordion_{}".format(index)):
                                        with tag("div", klass="card"):
                                            with tag(
                                                "a",
                                                ("data-toggle", "collapse"),
                                                ("data-target", "#job_list___sub_collapse_{}_1".format(index)),
                                                ("aria-expanded", "false"),
                                                ("aria-controls", "job_list___sub_collapse_{}_1".format(index)),
                                                href="#",
                                                klass="job_list___sub_multi_heading",
                                            ):
                                                with tag(
                                                    "div",
                                                    ("role", "alert"),
                                                    klass="card-header alert-{}".format(level),
                                                    id="job_list___sub_heading_{}_1".format(index),
                                                ):
                                                    with tag("button", klass="btn btn-outline-{}".format(level)):
                                                        with tag("span", klass="job_id"):
                                                            text("{} : Job ID {}".format(index, work_item.job_id))

                                            with tag(
                                                "div",
                                                ("aria-labelledby", "job_list___sub_heading_{}_1".format(index)),
                                                ("data-parent", "#job_list___sub_accordion_{}".format(index)),
                                                klass="collapse job_list___sub_multi_collapse",
                                                id="job_list___sub_collapse_{}_1".format(index),
                                            ):
                                                with tag("div", klass="card-body"):
                                                    with tag("div", klass="work-item"):
                                                        with tag(
                                                            "div",
                                                            klass="alert alert-{} test-outcome".format(level),
                                                            role="alert",
                                                        ):
                                                            if not result.is_killed:
                                                                with tag("p"):
                                                                    text("SURVIVED")
                                                            with tag("p"):
                                                                text("worker outcome: {}".format(result.worker_outcome))
                                                            with tag("p"):
                                                                text("test outcome: {}".format(result.test_outcome))

                                                    with tag("pre", klass="location"):
                                                        with tag(
                                                            "a",
                                                            href=pycharm_url(
                                                                str(work_item.module_path), work_item.start_pos[0]
                                                            ),
                                                            klass="text-secondary",
                                                        ):
                                                            with tag("button", klass="btn btn-outline-dark"):
                                                                text(
                                                                    "{}, start pos: {}, end pos: {}".format(
                                                                        work_item.module_path,
                                                                        work_item.start_pos,
                                                                        work_item.end_pos,
                                                                    )
                                                                )

                                                    with tag("pre"):
                                                        text(
                                                            "operator: {}, occurrence: {}".format(
                                                                work_item.operator_name, work_item.occurrence
                                                            )
                                                        )

                                                    if result is not None:
                                                        if result.diff:
                                                            with tag("div", klass="alert alert-secondary"):
                                                                with tag("pre", klass="diff"):
                                                                    text(result.diff)

                                                        if result.output:
                                                            with tag("div", klass="alert alert-secondary"):
                                                                with tag("pre", klass="diff"):
                                                                    text(result.output)

            with tag("script"):
                doc.attr(src="https://code.jquery.com/jquery-3.3.1.slim.min.js")
                doc.attr(("integrity", "sha384-q8i/X+965DzO0rT7abK41JStQIAqVgRVzpbzo5smXKp4YfRvH+8abtTE1Pi6jizo"))
                doc.attr(("crossorigin", "anonymous"))
            with tag("script"):
                doc.attr(src="https://cdnjs.cloudflare.com/ajax/libs/popper.js/1.14.7/umd/popper.min.js")
                doc.attr(("integrity", "sha384-UO2eT0CpHqdSJQ6hJty5KVphtPhzWj9WO1clHTMGa3JDZwrnQq4sF86dIHNDz0W1"))
                doc.attr(("crossorigin", "anonymous"))
            with tag("script"):
                doc.attr(src="https://stackpath.bootstrapcdn.com/bootstrap/4.3.1/js/bootstrap.min.js")
                doc.attr(("integrity", "sha384-JjSmVgyd0p3pXB1rRibZUAYoIIy6OrQ6VrjIEaFf/nJGzIxFDsf4x0xIM+B07jRM"))
                doc.attr(("crossorigin", "anonymous"))
            with tag("script", type="text/javascript"):
                doc.asis(
                    "$('div.job_list___sub_multi_collapse').on('shown.bs.collapse',"
                    "  function () {"
                    "    correct_behavior_functional_buttons();"
                    "  });"
                    "$('div.job_list___sub_multi_collapse').on('hidden.bs.collapse',"
                    "  function () {"
                    "    correct_behavior_functional_buttons();"
                    "  });"
                    "function correct_behavior_functional_buttons() {"
                    "    var expand = false;"
                    "    var collapse = false;"
                    "    $('a.job_list___sub_multi_heading').each(function(index) {"
                    "      if ($(this).attr('aria-expanded') == 'false') {"
                    "        expand = true;"
                    "        return false;"
                    "      };"
                    "    });"
                    "    $('a.job_list___sub_multi_heading').each(function(index) {"
                    "      if ($(this).attr('aria-expanded') == 'true') {"
                    "        collapse = true;"
                    "        return false;"
                    "      };"
                    "    });"
                    "    if (expand) {"
                    "      $('div#job_item_expand_all').css('display', 'inline-block');"
                    "    } else {"
                    "      $('div#job_item_expand_all').css('display', 'none');"
                    "    };"
                    "    if (collapse) {"
                    "      $('div#job_item_collapse_all').css('display', 'inline-block');"
                    "    } else {"
                    "      $('div#job_item_collapse_all').css('display', 'none');"
                    "    };"
                    "  };"
                    "correct_behavior_functional_buttons();"
                )

    return doc


def pycharm_url(filename, line_number):
    "Get a URL for opening a file in Pycharm."
    return "pycharm://open?file={}&line={}".format(filename, line_number)



"Base distributor implementation details."

import abc


class Distributor(metaclass=abc.ABCMeta):
    "Base class for work distribution strategies."

    @abc.abstractmethod
    def __call__(self, pending_work, test_command, timeout, distributor_config, on_task_complete):
        """Execute jobs in `pending_work_items`.

        Spend no more than `timeout` seconds for a single job, using `distributor_config` to
        distribute the work.
        """



"Implementation of the exception-replacement operator."

from parso.python.tree import Name, PythonNode

from cosmic_ray.exceptions import CosmicRayTestingException

from .operator import Operator


class ExceptionReplacer(Operator):
    """An operator that modifies exception handlers."""

    def mutation_positions(self, node):
        if isinstance(node, PythonNode):
            if node.type == "except_clause":
                for name in self._name_nodes(node):
                    yield (name.start_pos, name.end_pos)

    def mutate(self, node, index):
        assert isinstance(node, PythonNode)
        assert node.type == "except_clause"

        name_nodes = self._name_nodes(node)
        assert index < len(name_nodes)
        name_nodes[index].value = CosmicRayTestingException.__name__
        return node

    @staticmethod
    def _name_nodes(node):
        if isinstance(node.children[1], Name):
            return (node.children[1],)

        atom = node.children[1]
        test_list = atom.children[1]
        return test_list.children[::2]

    @classmethod
    def examples(cls):
        return (
            (
                "try: raise OSError\nexcept OSError: pass",
                "try: raise OSError\nexcept {}: pass".format(CosmicRayTestingException.__name__),
            ),
            (
                "try: raise OSError\nexcept (OSError, ValueError): pass",
                "try: raise OSError\nexcept (OSError, {}): pass".format(CosmicRayTestingException.__name__),
                1,
            ),
            (
                "try: raise OSError\nexcept (OSError, ValueError, KeyError): pass",
                "try: raise OSError\nexcept (OSError, {}, KeyError): pass".format(CosmicRayTestingException.__name__),
                1,
            ),
            ("try: pass\nexcept: pass", "try: pass\nexcept: pass"),
        )



"Tool for printing reports on mutation testing sessions."

import click

from cosmic_ray.work_db import use_db, WorkDB
from cosmic_ray.tools.survival_rate import kills_count, survival_rate


@click.command()
@click.option("--show-output/--no-show-output", default=False, help="Display output of test executions")
@click.option("--show-diff/--no-show-diff", default=False, help="Display diff of mutants")
@click.option("--show-pending/--no-show-pending", default=False, help="Display results for incomplete tasks")
@click.argument("session-file", type=click.Path(dir_okay=False, readable=True, exists=True))
def report(show_output, show_diff, show_pending, session_file):
    """Print a nicely formatted report of test results and some basic statistics."""

    with use_db(session_file, WorkDB.Mode.open) as db:
        for work_item, result in db.completed_work_items:
            display_work_item(work_item)

            print("worker outcome: {}, test outcome: {}".format(result.worker_outcome, result.test_outcome))

            if show_output:
                print("=== OUTPUT ===")
                print(result.output)
                print("==============")

            if show_diff:
                print("=== DIFF ===")
                print(result.diff)
                print("============")

        if show_pending:
            for work_item in db.pending_work_items:
                display_work_item(work_item)

        num_items = db.num_work_items
        num_complete = db.num_results

        print("total jobs: {}".format(num_items))

        if num_complete > 0:
            print("complete: {} ({:.2f}%)".format(num_complete, num_complete / num_items * 100))
            num_killed = kills_count(db)
            print("surviving mutants: {} ({:.2f}%)".format(num_complete - num_killed, survival_rate(db)))
        else:
            print("no jobs completed")


def display_work_item(work_item):
    print("[job-id] {}".format(work_item.job_id))
    for mutation in work_item.mutations:
        print("{} {} {}".format(mutation.module_path, mutation.operator_name, mutation.occurrence))


if __name__ == "__main__":
    report()  # no-qa: no-value-for-parameter


"""Implementation of the 'new-config' command.
"""

import os.path

import qprompt

from cosmic_ray.config import ConfigDict
from cosmic_ray.plugins import distributor_names


MODULE_PATH_HELP = """The path to the module that will be mutated.

If this is a package (as opposed to a single file module),
then all modules in the package and its subpackages will be
mutated.

This path can be absolute or relative to the location of the
config file.
"""


TEST_COMMAND_HELP = """The command to execute to run the tests on mutated code.
"""


def new_config():
    """Prompt user for config variables and generate new config.

    Returns: A new ConfigDict.
    """
    config = ConfigDict()
    config["module-path"] = qprompt.ask_str(
        "Top-level module path", blk=False, vld=os.path.exists, hlp=MODULE_PATH_HELP
    )

    timeout = qprompt.ask_str(
        "Test execution timeout (seconds)",
        vld=float,
        blk=False,
        hlp="The number of seconds to let a test run before terminating it.",
    )
    config["timeout"] = float(timeout)
    config["excluded-modules"] = []

    config["test-command"] = qprompt.ask_str("Test command", blk=False, hlp=TEST_COMMAND_HELP)

    menu = qprompt.Menu()
    for at_pos, distributor_name in enumerate(distributor_names()):
        menu.add(str(at_pos), distributor_name)
    config["distributor"] = ConfigDict()
    config["distributor"]["name"] = menu.show(header="Distributor", returns="desc")

    return config



"Support for running tests in a subprocess."

import asyncio
import logging
import os
import traceback

from cosmic_ray.work_item import TestOutcome

log = logging.getLogger(__name__)

# We use an asyncio-subprocess-based approach here instead of a simple
# subprocess.run()-based approach because there are problems with timeouts and
# reading from stderr in subprocess.run. Since we have to be prepared for test
# processes that run longer than timeout (and, indeed, which run forever), the
# broken subprocess stuff simply doesn't work. So we do this, which seesm to
# work on all platforms.


async def run_tests(command, timeout):
    """Run test command in a subprocess.

    If the command exits with status 0, then we assume that all tests passed. If
    it exits with any other code, we assume a test failed. If the call to launch
    the subprocess throws an exception, we consider the test 'incompetent'.

    Tests which time out are considered 'killed' as well.

    Args:
        command (str): The command to execute.
        timeout (number): The maximum number of seconds to allow the tests to run.

    Return: A tuple `(TestOutcome, output)` where the `output` is a string
        containing the output of the command.
    """
    log.info("Running test (timeout=%s): %s", timeout, command)

    # We want to avoid writing pyc files in case our changes happen too fast for Python to
    # notice them. If the timestamps between two changes are too small, Python won't recompile
    # the source.
    env = dict(os.environ)
    env["PYTHONDONTWRITEBYTECODE"] = "1"

    try:
        proc = await asyncio.create_subprocess_shell(
            command, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.STDOUT, env=env
        )
    except Exception:  # pylint: disable=W0703
        return (TestOutcome.INCOMPETENT, traceback.format_exc())

    try:
        outs, _errs = await asyncio.wait_for(proc.communicate(), timeout)

        assert proc.returncode is not None

        if proc.returncode == 0:
            return (TestOutcome.SURVIVED, outs.decode("utf-8"))
        else:
            return (TestOutcome.KILLED, outs.decode("utf-8"))

    except asyncio.TimeoutError:
        proc.terminate()
        return (TestOutcome.KILLED, "timeout")

    except Exception:  # pylint: disable=W0703
        proc.terminate()
        return (TestOutcome.INCOMPETENT, traceback.format_exc())

    finally:
        await proc.wait()



"Implementation of the 'execute' command."
import logging
import os

from cosmic_ray.config import ConfigDict
from cosmic_ray.plugins import get_distributor
from cosmic_ray.progress import reports_progress

log = logging.getLogger(__name__)

_progress_messages = {}  # pylint: disable=invalid-name


def _update_progress(work_db):
    num_work_items = work_db.num_work_items
    pending = num_work_items - work_db.num_results
    total = num_work_items
    remaining = total - pending
    message = "{} out of {} completed".format(remaining, total)
    _progress_messages[work_db.name] = message


def _report_progress(stream):
    for db_name, progress_message in _progress_messages.items():
        session = os.path.splitext(db_name)[0]
        print("{session} : {progress_message}".format(session=session, progress_message=progress_message), file=stream)


@reports_progress(_report_progress)
def execute(work_db, config: ConfigDict):
    """Execute any pending work in the database `work_db`,
    recording the results.

    This looks for any work in `work_db` which has no results, schedules it to
    be executed, and records any results that arrive.
    """
    _update_progress(work_db)
    distributor = get_distributor(config.distributor_name)

    def on_task_complete(job_id, work_result):
        work_db.set_result(job_id, work_result)
        _update_progress(work_db)
        log.info("Job %s complete", job_id)

    log.info("Beginning execution")
    distributor(
        work_db.pending_work_items,
        config.test_command,
        config.timeout,
        config.distributor_config,
        on_task_complete=on_task_complete,
    )
    log.info("Execution finished")



"""A filter that uses "# pragma: no mutate" to determine when specific mutations
should be skipped.
"""
import logging
import re
import sys
from functools import lru_cache

from cosmic_ray.tools.filters.filter_app import FilterApp
from cosmic_ray.work_item import WorkerOutcome, WorkResult

log = logging.getLogger()


class PragmaNoMutateFilter(FilterApp):
    """Implements the pragma-no-mutate filter."""

    def description(self):
        return __doc__

    def filter(self, work_db, _args):
        """Mark lines with "# pragma: no mutate" as SKIPPED

        For all work_item in db, if the LAST line of the working zone is marked
        with "# pragma: no mutate", This work_item will be skipped.
        """

        @lru_cache()
        def file_contents(file_path):
            "A simple cache of file contents."
            with file_path.open(mode="rt") as handle:
                return handle.readlines()

        re_is_mutate = re.compile(r".*#.*pragma:.*no mutate.*")

        for item in work_db.work_items:
            for mutation in item.mutations:
                print(mutation.module_path)
                lines = file_contents(mutation.module_path)
                try:
                    # item.{start,end}_pos[0] seems to be 1-based.
                    line_number = mutation.end_pos[0] - 1
                    if mutation.end_pos[1] == 0:
                        # The working zone ends at begin of line,
                        # consider the previous line.
                        line_number -= 1
                    line = lines[line_number]
                    if re_is_mutate.match(line):
                        work_db.set_result(
                            item.job_id,
                            WorkResult(output=None, test_outcome=None, diff=None, worker_outcome=WorkerOutcome.SKIPPED),
                        )
                except Exception as ex:
                    raise Exception(
                        "module_path: %s, start_pos: %s, end_pos: %s, len(lines): %s"
                        % (mutation.module_path, mutation.start_pos, mutation.end_pos, len(lines))
                    ) from ex


def main(argv=None):
    """Run pragma-no-mutate filter with specified command line arguments."""
    return PragmaNoMutateFilter().main(argv)


if __name__ == "__main__":
    sys.exit(main())



"""Operator-provider plugin for all core cosmic ray operators.
"""

import itertools

from . import (binary_operator_replacement, boolean_replacer, break_continue,
               comparison_operator_replacement, exception_replacer,
               no_op, number_replacer, remove_decorator, unary_operator_replacement,
               zero_iteration_for_loop)

# NB: The no_op operator gets special handling. We don't include it in iteration of the
# available operators. However, you can request it from the provider by name. This lets us
# use it in a special way: to request that a worker perform a no-op test run while preventing
# it from being used in normal mutations testing runs.

_OPERATORS = {
    op.__name__: op
    for op in itertools.chain(binary_operator_replacement.operators(
    ), comparison_operator_replacement.operators(
    ), unary_operator_replacement.operators(), (
        boolean_replacer.AddNot, boolean_replacer.ReplaceTrueWithFalse,
        boolean_replacer.ReplaceFalseWithTrue,
        boolean_replacer.ReplaceAndWithOr, boolean_replacer.ReplaceOrWithAnd,
        break_continue.ReplaceBreakWithContinue,
        break_continue.ReplaceContinueWithBreak,
        exception_replacer.ExceptionReplacer,
        number_replacer.NumberReplacer,
        remove_decorator.RemoveDecorator,
        zero_iteration_for_loop.ZeroIterationForLoop))
}


class OperatorProvider:
    """Provider for all of the core Cosmic Ray operators."""

    def __iter__(self):
        return iter(_OPERATORS)

    def __getitem__(self, name):
        if name == 'NoOp':
            return no_op.NoOp

        return _OPERATORS[name]



"Tools for working with parso ASTs."

from abc import ABC, abstractmethod
import io
import sys

import parso.python.tree
import parso.tree


class Visitor(ABC):
    """AST visitor for parso trees.

    This supports both simple traversal as well as editing of the tree.
    """

    def walk(self, node):
        "Walk a parse tree, calling visit for each node."
        node = self.visit(node)

        if node is None:
            return None

        if isinstance(node, parso.tree.BaseNode):
            walked = map(self.walk, node.children)
            node.children = [child for child in walked if child is not None]

        return node

    @abstractmethod
    def visit(self, node):
        """Called for each node in the walk.

        This should return a node that will replace the node argument in the AST. This can be
        the node argument itself, a new node, or None. If None is returned, then the node is
        removed from the tree.

        Args:
            node: The node currently being visited.

        Returns:
            A node or `None`.
        """


def ast_nodes(node):
    """Iterable of all nodes in a tree.

    Args:
        node: The top node in a parso tree to iterate.

    Yields:
        All of the nodes in the tree.
    """
    yield node

    if isinstance(node, parso.tree.BaseNode):
        for child in node.children:
            yield from ast_nodes(child)


def get_ast(module_path):
    """Get the AST for the code in a file.

    Args:
        module_path: pathlib.Path to the file containing the code.

    Returns:
        The parso parse tree for the code in `module_path`.
    """
    with module_path.open(mode="rt", encoding="utf-8") as handle:
        source = handle.read()

    return parso.parse(source)


def is_none(node):
    "Determine if a node is the `None` keyword."
    return isinstance(node, parso.python.tree.Keyword) and node.value == "None"


def is_number(node):
    "Determine if a node is a number."
    return isinstance(node, parso.python.tree.Number)


def dump_node(node):
    "Generate string version of node."
    buffer = io.StringIO()
    write = buffer.write

    def do_dump(node, indent=""):
        write("{}{}({}".format(indent, type(node).__name__, node.type))
        value = getattr(node, "value", None)
        if value:
            value = value.replace("\n", "\\n")
            write(", '{}'".format(value))
        children = getattr(node, "children", None)
        if children:
            write(", [\n")
            for child in children:
                do_dump(child, indent + " " * 4)
                write(",\n")
            write("{}]".format(indent))
        write(")")
        if not indent:
            write("\n")

    do_dump(node)
    return buffer.getvalue()



"Implementation of the remove-decorator operator."

from parso.python.tree import Decorator

from .operator import Operator


class RemoveDecorator(Operator):
    """An operator that removes decorators."""

    def mutation_positions(self, node):
        if isinstance(node, Decorator):
            yield (node.start_pos, node.end_pos)

    def mutate(self, node, index):
        assert isinstance(node, Decorator)
        assert index == 0

    @classmethod
    def examples(cls):
        return (
            ('@foo\ndef bar(): pass', 'def bar(): pass'),
            ('@first\n@second\ndef bar(): pass', '@second\ndef bar(): pass'),
            ('@first\n@second\ndef bar(): pass', '@first\ndef bar(): pass', 1),
            ('@first\n@second\n@third\ndef bar(): pass', '@first\n@third\ndef bar(): pass', 1),
        )



"""Implementation of the unary-operator-replacement operator.
"""

from enum import Enum
from itertools import permutations

from parso.python.tree import Keyword, Operator, PythonNode

from . import operator
from .util import extend_name


class UnaryOperators(Enum):
    "All unary operators that we mutate."
    UAdd = '+'
    USub = '-'
    Invert = '~'
    Not = 'not '
    Nothing = None


def _create_replace_unary_operators(from_op, to_op):
    if to_op.value is None:
        suffix = '_Delete_{}'.format(from_op.name)
    else:
        suffix = '_{}_{}'.format(from_op.name, to_op.name)

    @extend_name(suffix)
    class ReplaceUnaryOperator(operator.Operator):
        "An operator that replaces unary {} with unary {}.".format(
            from_op.name, to_op.name)

        def mutation_positions(self, node):
            if _is_unary_operator(node):
                op = node.children[0]
                if op.value.strip() == from_op.value.strip():
                    yield (op.start_pos, op.end_pos)

        def mutate(self, node, index):
            assert index == 0
            assert _is_unary_operator(node)

            if to_op.value is None:
                # This is a bit goofy since it can result in "return not x"
                # becoming "return  x" (i.e. with two spaces). But it's correct
                # enough.
                node.children[0].value = ''
            else:
                node.children[0].value = to_op.value
            return node

        @classmethod
        def examples(cls):
            from_code = '{}1'.format(from_op.value)
            to_code = from_code[len(from_op.value):]

            if to_op is not UnaryOperators.Nothing:
                to_code = to_op.value + to_code
            elif from_op is UnaryOperators.Not:
                to_code = ' ' + to_code

            return (
                (from_code, to_code),
            )

    return ReplaceUnaryOperator


def _is_factor(node):
    return (isinstance(node, PythonNode)
            and node.type in {'factor', 'not_test'} and len(node.children) > 0
            and isinstance(node.children[0], Operator))


def _is_not_test(node):
    return (isinstance(node, PythonNode) and node.type == 'not_test'
            and len(node.children) > 0
            and isinstance(node.children[0], Keyword)
            and node.children[0].value == 'not')


def _is_unary_operator(node):
    return _is_factor(node) or _is_not_test(node)


def _prohibited(from_op, to_op):
    "Determines if from_op is allowed to be mutated to to_op."
    # 'not' can only be removed but not replaced with
    # '+', '-' or '~' b/c that may lead to strange results
    if from_op is UnaryOperators.Not:
        if to_op is not UnaryOperators.Nothing:
            return True

    # '+1' => '1' yields equivalent mutations
    if from_op is UnaryOperators.UAdd:
        if to_op is UnaryOperators.Nothing:
            return True

    return False


_MUTATION_OPERATORS = tuple(
    _create_replace_unary_operators(from_op, to_op)
    for (from_op, to_op) in permutations(UnaryOperators, 2)
    if from_op.value is not None if not _prohibited(from_op, to_op))

for op_cls in _MUTATION_OPERATORS:
    globals()[op_cls.__name__] = op_cls


def operators():
    "Iterable of unary operator mutation operators."
    return _MUTATION_OPERATORS



"Implementation of the no-op operator."

from .operator import Operator


class NoOp(Operator):
    """An operator that makes no changes.

    This is primarily for baselining and debugging. It behaves like any other operator, but it makes no changes.
    Obviously this means that, if your test suite passes on unmutated code, it will still pass after applying this
    operator. Use with care.
    """

    def mutation_positions(self, node):
        yield (node.start_pos, node.end_pos)

    def mutate(self, node, index):
        return node

    @classmethod
    def examples(cls):
        return (
            ('@foo\ndef bar(): pass', '@foo\ndef bar(): pass'),
            ('def bar(): pass', 'def bar(): pass'),
            ('1 + 1', '1 + 1'),
        )



"""Implementation of the replace-break-with-continue and
replace-continue-with-break operators.
"""

from .keyword_replacer import KeywordReplacementOperator


class ReplaceBreakWithContinue(KeywordReplacementOperator):
    "Operator which replaces 'break' with 'continue'."
    from_keyword = 'break'
    to_keyword = 'continue'


class ReplaceContinueWithBreak(KeywordReplacementOperator):
    "Operator which replaces 'continue' with 'break'."
    from_keyword = 'continue'
    to_keyword = 'break'
